const util = require('util');
const fs = require('fs');
const path = require('path');

const mkdirp = util.promisify(require('mkdirp'));
const rimraf = util.promisify(require('rimraf'));
const copyFile = util.promisify(fs.copyFile);
const unlink = util.promisify(fs.unlink);
const rmdir = util.promisify(fs.rmdir);

const glob = require('glob');

const concat = (x,y) =>
  x.concat(y)

const flatMap = (f,xs) =>
  xs.map(f).reduce(concat, [])

Array.prototype.flatMap = function(f) {
  return flatMap(f,this)
}

const defaultOption = {
  overwrite: true,
  removeSource: false,
  removeEmptyDirectory: false,
  removeDeletedFiles: false,
};

let sync = (source, destination, patterns) => {
  source = path.resolve(source);
  destination = path.resolve(destination);

  // パターン指示を正規化
  patterns = (patterns || ["**/*"])
  .map((pattern, index) => {
    if (typeof(pattern) == 'object' && pattern.pattern) {
      return {
        pattern: pattern.pattern,
        overwrite: pattern.hasOwnProperty('overwrite') ? pattern.overwrite : defaultOption.overwrite,
        removeSource: pattern.hasOwnProperty('removeSource') ? pattern.removeSource : defaultOption.removeSource,
        removeEmptyDirectory: pattern.hasOwnProperty('removeEmptyDirectory') ? pattern.removeEmptyDirectory : defaultOption.removeEmptyDirectory,
        removeDeletedFiles: pattern.hasOwnProperty('removeDeletedFiles') ? pattern.removeDeletedFiles : defaultOption.removeDeletedFiles,
      };
    }
    if (typeof(pattern) == 'string') {
      return {
        pattern: pattern,
        overwrite: defaultOption.overwrite,
        removeSource: defaultOption.removeSource,
        removeEmptyDirectory: defaultOption.removeEmptyDirectory,
        removeDeletedFiles: defaultOption.removeDeletedFiles,
      };
    }
  });
  // 対象ファイル情報一覧を構築
  let sourceFileList = patterns
  // パターンをファイル・ディレクトリの一覧に展開
  .flatMap(
    (pattern, index) => glob
    .sync(path.join(source, pattern.pattern))
    .map((entry) => {
      let filePath = entry.replace(new RegExp('^' + source + '/'), '');
      return {
        path: filePath,
        sourcePath: path.join(source, filePath),
        destinationPath: path.join(destination, filePath),
        overwrite: pattern.overwrite,
        removeSource: pattern.removeSource,
        removeEmptyDirectory: pattern.removeEmptyDirectory,
        removeDeletedFiles: pattern.removeDeletedFiles,
      };
    })
  )
  // ディレクトリの処理を最後に回したいので逆順ソート
  .sort((a, b) => {
    if (a.sourcePath < b.sourcePath) {
      return 1;
    }
    if (a.sourcePath > b.sourcePath) {
      return -1;
    }
    return 0;
  });
  // 対象ファイル情報一覧を構築
  let destinationFileList = patterns
  // パターンをファイル・ディレクトリの一覧に展開
  .flatMap(
    (pattern, index) => glob
    .sync(path.join(destination, pattern.pattern))
    .map((entry) => {
      let filePath = entry.replace(new RegExp('^' + destination + '/'), '');
      return {
        path: filePath,
        sourcePath: path.join(source, filePath),
        destinationPath: path.join(destination, filePath),
        overwrite: pattern.overwrite,
        removeSource: pattern.removeSource,
        removeEmptyDirectory: pattern.removeEmptyDirectory,
        removeDeletedFiles: pattern.removeDeletedFiles,
      };
    })
  )
  // ディレクトリの処理を最後に回したいので逆順ソート
  .sort((a, b) => {
    if (a.sourcePath < b.sourcePath) {
      return 1;
    }
    if (a.sourcePath > b.sourcePath) {
      return -1;
    }
    return 0;
  });

  return Promise
  .resolve()
  // コピー
  .then(
    () => sourceFileList
    // 配列の全要素を順次実行させるための Promise に変換
    .map((entry) => () => new Promise(
      (resolve, reject) => Promise
      .resolve()
      // 宛先ディレクトリ作成
      .then(() => mkdirp(path.dirname(entry.destinationPath)))
      // ファイルコピー
      .then(() => {
        if (
          !fs.statSync(entry.sourcePath).isDirectory()
          && (entry.overwrite || !fs.existsSync(entry.destinationPath))
        ) {
          return copyFile(entry.sourcePath, entry.destinationPath);
        }
        return Promise.resolve(entry);
      })
      // 元ファイル削除（オプション）
      .then(() => {
        // 通常ファイルを削除
        if (
          entry.removeSource
          && !fs.statSync(entry.sourcePath).isDirectory()
        ) {
          return unlink(entry.sourcePath);
        }
        // 空ディレクトリを削除（オプション）
        if (
          entry.removeEmptyDirectory
          && fs.statSync(entry.sourcePath).isDirectory()
          && !fs.readdirSync(entry.sourcePath).some((item) => !/^\.+$/.test(item))
        ) {
          return rmdir(entry.sourcePath);
        }
        return Promise.resolve(entry);
      })
      .then(resolve)
    ))
    // 配列の一つ前の要素が終わったら、現在の要素を実行する
    .reduce(
      (previous, current, index, original) => previous.then(current),
      Promise.resolve(sourceFileList[0])
    )
  )
  // 削除済みファイルの削除（オプション）
  .then(
    () => destinationFileList
    // 削除フラグが立っており、元ファイルリストに存在しない要素に絞り込み
    .filter((item) => item.removeDeletedFiles && !sourceFileList.some((x) => x.path == item.path))
    // 削除済みファイルの削除
    .map((entry) => () => new Promise(
      (resolve, reject) => Promise
      .resolve()
      .then(() => {
        // 通常ファイルを削除
        if (
          !fs.statSync(entry.destinationPath).isDirectory()
        ) {
          return unlink(entry.destinationPath);
        }
        // 空ディレクトリを削除
        if (
          fs.statSync(entry.destinationPath).isDirectory()
          && !fs.readdirSync(entry.destinationPath).some((item) => !/^\.+$/.test(item))
        ) {
          return rmdir(entry.destinationPath);
        }
        return Promise.resolve(entry);
      })
      .then(resolve)
    ))
    .reduce(
      (previous, current, index, original) => previous.then(current),
      Promise.resolve(destinationFileList[0])
    )
  );
};

module.exports = sync;

